"use strict";
// Copyright 2020 Google Inc. Use of this source code is governed by an
// MIT-style license that can be found in the LICENSE file or at
// https://opensource.org/licenses/MIT.
Object.defineProperty(exports, "__esModule", { value: true });
exports.compileStringAsync = exports.compileAsync = exports.compileString = exports.compile = void 0;
const p = require("path");
const supportsColor = require("supports-color");
const proto = require("./vendor/embedded-protocol/embedded_sass_pb");
const utils = require("./utils");
const async_compiler_1 = require("./async-compiler");
const dispatcher_1 = require("./dispatcher");
const exception_1 = require("./exception");
const function_registry_1 = require("./function-registry");
const message_transformer_1 = require("./message-transformer");
const packet_transformer_1 = require("./packet-transformer");
const sync_compiler_1 = require("./sync-compiler");
function compile(path, options) {
    // TODO(awjin): Create logger, importer, function registries.
    return compileRequestSync(newCompilePathRequest(path, options), options);
}
exports.compile = compile;
function compileString(source, options) {
    // TODO(awjin): Create logger, importer, function registries.
    return compileRequestSync(newCompileStringRequest(source, options), options);
}
exports.compileString = compileString;
function compileAsync(path, options) {
    // TODO(awjin): Create logger, importer, function registries.
    return compileRequestAsync(newCompilePathRequest(path, options), options);
}
exports.compileAsync = compileAsync;
function compileStringAsync(source, options) {
    // TODO(awjin): Create logger, importer, function registries.
    return compileRequestAsync(newCompileStringRequest(source, options), options);
}
exports.compileStringAsync = compileStringAsync;
// Creates a request for compiling a file.
function newCompilePathRequest(path, options) {
    // TODO(awjin): Populate request with importer/function IDs.
    const request = newCompileRequest(options);
    request.setPath(path);
    return request;
}
// Creates a request for compiling a string.
function newCompileStringRequest(source, options) {
    // TODO(awjin): Populate request with importer/function IDs.
    var _a;
    const input = new proto.InboundMessage.CompileRequest.StringInput();
    input.setSource(source);
    switch ((_a = options === null || options === void 0 ? void 0 : options.syntax) !== null && _a !== void 0 ? _a : 'scss') {
        case 'scss':
            input.setSyntax(proto.Syntax.SCSS);
            break;
        case 'indented':
            input.setSyntax(proto.Syntax.INDENTED);
            break;
        case 'css':
            input.setSyntax(proto.Syntax.CSS);
            break;
        default:
            throw new Error(`Unknown options.syntax: "${options === null || options === void 0 ? void 0 : options.syntax}"`);
    }
    if (options === null || options === void 0 ? void 0 : options.url)
        input.setUrl(options.url.toString());
    const request = newCompileRequest(options);
    request.setString(input);
    return request;
}
// Creates a compilation request for the given `options` without adding any
// input-specific options.
function newCompileRequest(options) {
    var _a, _b, _c, _d;
    const request = new proto.InboundMessage.CompileRequest();
    request.setGlobalFunctionsList(Object.keys((_a = options === null || options === void 0 ? void 0 : options.functions) !== null && _a !== void 0 ? _a : {}));
    request.setSourceMap(!!(options === null || options === void 0 ? void 0 : options.sourceMap));
    request.setAlertColor((_b = options === null || options === void 0 ? void 0 : options.alertColor) !== null && _b !== void 0 ? _b : !!supportsColor.stdout);
    request.setAlertAscii(!!(options === null || options === void 0 ? void 0 : options.alertAscii));
    request.setQuietDeps(!!(options === null || options === void 0 ? void 0 : options.quietDeps));
    request.setVerbose(!!(options === null || options === void 0 ? void 0 : options.verbose));
    switch ((_c = options === null || options === void 0 ? void 0 : options.style) !== null && _c !== void 0 ? _c : 'expanded') {
        case 'expanded':
            request.setStyle(proto.OutputStyle.EXPANDED);
            break;
        case 'compressed':
            request.setStyle(proto.OutputStyle.COMPRESSED);
            break;
        default:
            throw new Error(`Unknown options.style: "${options === null || options === void 0 ? void 0 : options.style}"`);
    }
    for (const path of (_d = options === null || options === void 0 ? void 0 : options.loadPaths) !== null && _d !== void 0 ? _d : []) {
        const importer = new proto.InboundMessage.CompileRequest.Importer();
        importer.setPath(p.resolve(path));
        request.addImporters(importer);
    }
    return request;
}
// Spins up a compiler, then sends it a compile request. Returns a promise that
// resolves with the CompileResult. Throws if there were any protocol or
// compilation errors. Shuts down the compiler after compilation.
async function compileRequestAsync(request, options) {
    const functionRegistry = new function_registry_1.FunctionRegistry(options === null || options === void 0 ? void 0 : options.functions);
    const embeddedCompiler = new async_compiler_1.AsyncEmbeddedCompiler();
    try {
        // TODO(awjin): Pass import and function registries' handler functions to
        // dispatcher.
        const dispatcher = createDispatcher(embeddedCompiler.stdout$, buffer => {
            embeddedCompiler.writeStdin(buffer);
        }, {
            handleImportRequest: () => {
                throw Error('Custom importers not yet implemented.');
            },
            handleFileImportRequest: () => {
                throw Error('Custom file importers not yet implemented.');
            },
            handleCanonicalizeRequest: () => {
                throw Error('Canonicalize not yet implemented.');
            },
            handleFunctionCallRequest: request => functionRegistry.call(request),
        });
        // TODO(awjin): Subscribe logger to dispatcher's log events.
        return handleCompileResponse(await new Promise((resolve, reject) => dispatcher.sendCompileRequest(request, (err, response) => {
            if (err) {
                reject(err);
            }
            else {
                resolve(response);
            }
        })));
    }
    finally {
        embeddedCompiler.close();
        await embeddedCompiler.exit$;
    }
}
// Spins up a compiler, then sends it a compile request. Returns a promise that
// resolves with the CompileResult. Throws if there were any protocol or
// compilation errors. Shuts down the compiler after compilation.
function compileRequestSync(request, options) {
    const functionRegistry = new function_registry_1.FunctionRegistry(options === null || options === void 0 ? void 0 : options.functions);
    const embeddedCompiler = new sync_compiler_1.SyncEmbeddedCompiler();
    try {
        // TODO(awjin): Pass import and function registries' handler functions to
        // dispatcher.
        const dispatcher = createDispatcher(embeddedCompiler.stdout$, buffer => {
            embeddedCompiler.writeStdin(buffer);
        }, {
            handleImportRequest: () => {
                throw Error('Custom importers not yet implemented.');
            },
            handleFileImportRequest: () => {
                throw Error('Custom file importers not yet implemented.');
            },
            handleCanonicalizeRequest: () => {
                throw Error('Canonicalize not yet implemented.');
            },
            handleFunctionCallRequest: request => functionRegistry.call(request),
        });
        // TODO(awjin): Subscribe logger to dispatcher's log events.
        let error;
        let response;
        dispatcher.sendCompileRequest(request, (error_, response_) => {
            if (error_) {
                error = error_;
            }
            else {
                response = response_;
            }
        });
        for (;;) {
            if (!embeddedCompiler.yield()) {
                throw utils.compilerError('Embedded compiler exited unexpectedly.');
            }
            if (error)
                throw error;
            if (response)
                return handleCompileResponse(response);
        }
    }
    finally {
        embeddedCompiler.close();
        embeddedCompiler.yieldUntilExit();
    }
}
/**
 * Creates a dispatcher that dispatches messages from the given `stdout` stream.
 */
function createDispatcher(stdout, writeStdin, handlers) {
    const packetTransformer = new packet_transformer_1.PacketTransformer(stdout, writeStdin);
    const messageTransformer = new message_transformer_1.MessageTransformer(packetTransformer.outboundProtobufs$, packet => packetTransformer.writeInboundProtobuf(packet));
    return new dispatcher_1.Dispatcher(messageTransformer.outboundMessages$, message => messageTransformer.writeInboundMessage(message), handlers);
}
/**
 * Converts a `CompileResponse` into a `CompileResult`.
 *
 * Throws a `SassException` if the compilation failed.
 */
function handleCompileResponse(response) {
    if (response.getSuccess()) {
        const success = response.getSuccess();
        const result = {
            css: success.getCss(),
            loadedUrls: success.getLoadedUrlsList().map(url => new URL(url)),
        };
        const sourceMap = success.getSourceMap();
        if (sourceMap)
            result.sourceMap = JSON.parse(sourceMap);
        return result;
    }
    else if (response.getFailure()) {
        throw new exception_1.Exception(response.getFailure());
    }
    else {
        throw utils.compilerError('Compiler sent empty CompileResponse.');
    }
}
//# sourceMappingURL=compile.js.map